/*
 * gamehandler.c
 */

#include "gamehandler.h"
#include "table.h"
#include "callbacks.h"
#include "dice.h"
#include <stdlib.h>
#include <time.h>


GtkWidget *mainwidget;

void
init_gamehandler(GtkWidget *widget)
{
    int i;
    
    srand(time(0));

    g_assert(widget != NULL);
    mainwidget = widget;

    gamehandler.current_player = 1;
    gamehandler.player_num = 1;
    gamehandler.player_name[0] = "Player 1";
    gamehandler.player_name[1] = "Player 2";
    gamehandler.player_name[2] = "Player 3";
    gamehandler.gameflow_status = GAMEFLOW_UNDEFINED;
    for (i=0; i<MAX_PLAYERS; i++)
	gamehandler.player_is_done[i] = FALSE;
    gamehandler_reset_values();

    statusbar("With luck you will win.");
}

// Reset the gamehandler struct values, called after each turn
void
gamehandler_reset_values()
{
    int i;

    gamehandler.turn = 0;
    for (i=0; i<SLOTS; i++)
    {
	gamehandler.rolls[i] = i+1;
	gamehandler.rolls_stored[i] = 0;
    }
}

// Reset everything for a new game
void
gamehandler_new_game()
{
    int i;

    g_print("gamehandler new game\n");

    gamehandler_reset_values();
    gamehandler.gameflow_status = GAMEFLOW_UNDEFINED;
    gamehandler.current_player = 1;
    reset_table();
    for (i=0; i<5; i++)
    {
	dice_hilite(i, mainwidget);
    }
    for (i=0; i<MAX_PLAYERS; i++)
	gamehandler.player_is_done[i] = FALSE;

    statusbar("New game, new luck...");
}

// Returns a random number ranged from 1 to 6.
gint
gamehandler_get_random_number()
{
    return (int)((float)rand() / RAND_MAX * 6) + 1;
}

// Interface for callbacks. Here we accept the gamehandler part for
// the work when "Do Roll" button was pressed.
void
gamehandler_handle_roll(int *rolls, GtkWidget *parent)
{
    int i, counter = 0;
    
    for (i=0; i<SLOTS; i++)
    {
	if (!gamehandler.rolls_stored[i])
	{
	    gamehandler.rolls[i] = rolls[i];
	    counter ++;
	}
    }
    g_print("COUNTER: %d\n", counter);

    gamehandler.turn ++;
    gamehandler.gameflow_status = GAMEFLOW_ROLLING;
    g_print("Turn %d\n", gamehandler.turn);

    // 3 rolls done or all dice marked
    if (gamehandler.turn >= MAX_TURNS || !counter)
    {
	gamehandler_process_turn();
	gamehandler.turn = 0;
	// Print the possible values in the table
	table_update_column();
	// Change the game status
	gamehandler.gameflow_status = GAMEFLOW_CHOOSING;
	statusbar("Now choose a row.");
    }
}

// Interface for callbacks. Here we do our work when a die is clicked, check
// if we select or unselect it.
void
gamehandler_select_die(int which)
{
    g_assert(which >= 0 && which <= 5);

    if (gamehandler.rolls_stored[which])
	gamehandler.rolls_stored[which] = 0;
    else
	gamehandler.rolls_stored[which] = gamehandler.rolls[which];
}

// Quicksort implemention, needed below in gamehandler_process_turn
void
gamehandler_quicksort(int feld[], int anf, int ende)
{
    int l = anf, r = ende, tmp, pivot = feld[ende];

    do {
	while (feld[l] < pivot)
	    l++;

	while (feld[r] > pivot)
	    r--;

	if (l <= r)
	{
	    tmp = feld[l];
	    feld[l] = feld[r];
	    feld[r] = tmp;
			
	    l++;
	    r--;
	}
    } while (l <= r);

    if (anf < r)
	gamehandler_quicksort(feld, anf, r);
    if (l < ende)
	gamehandler_quicksort(feld, l, ende);
}

// Parsing of the dice values after a move is done, calculate the
// result and store it in the result struct.
void
gamehandler_process_turn()
{
    // Deep copy the array to a tmp array and fill unfilled
    // stored rolls slots with the left unmarked dice
    int i,
	j,
	counter = 0,
	rolls_tmp[SLOTS];

    // Clear result struct
    result.pasch = 0;
    result.erg = 0;
    for (i=0; i<6; i++)
	result.numbers[i] = 0;
	
    for (i=0; i<SLOTS; i++)
    {
	if (!gamehandler.rolls_stored[i])
	    rolls_tmp[i] = gamehandler.rolls[i];
	else
	    rolls_tmp[i] = gamehandler.rolls_stored[i];
    }

    // Sort the stored rolls_tmp field
    gamehandler_quicksort(rolls_tmp, 0, SLOTS-1);

    // Now parse the stored rolls_tmp field for patterns
    
    /////////////////////////////////
    // Count the numbers
    for (i=1; i<7; i++)
	for (j=0; j<SLOTS; j++)
	    if (rolls_tmp[j] == i)
		result.numbers[i-1] ++;
				
    /////////////////////////////////
    // Pasch
    for (i=0; i<6; i++)
    {
	if (result.numbers[i] == 3)
	    result.pasch = (i+1) | PASCH3_FLAG;
	else if (result.numbers[i] == 4)
	    result.pasch = (i+1) | PASCH4_FLAG;
	else if (result.numbers[i] == 5)
	    result.pasch = PASCH5_FLAG;
    }
    g_print("PASCH = %d\n", result.pasch);

    /////////////////////////////////
    // Full House
    for (i=0; i<6; i++)
    {
	if (result.numbers[i] == 2 && !counter)
	{
	    counter = 1;
	    continue;
	}
	if (result.numbers[i] == 3 && !counter)
	{
	    counter = 2;
	    continue;
	}
	if (result.numbers[i] == 3 && counter == 1)
	{
	    counter = 3;
	    break;
	}
	if (result.numbers[i] == 2 && counter == 2)
	{
	    counter = 3;
	    break;
	}
    }
    if (counter == 3)
    {
	g_print("FULL HOUSE\n");
	result.erg =  FULL_HOUSE;
    }

    /////////////////////////////////
    // Streets
    counter = 0;
    for (i=0; i<SLOTS-1; i++)  // First remove double values from the temp array
	if (rolls_tmp[i] == rolls_tmp[i+1])
	    rolls_tmp[i] = 0;
    gamehandler_quicksort(rolls_tmp, 0, SLOTS-1);
    for (i=0; i<SLOTS-1; i++)
	if (rolls_tmp[i] && rolls_tmp[i] == rolls_tmp[i+1] - 1)
	    counter ++;			
    if (counter == SLOTS - 1)
    {
	g_print("GRAND STREET\n");
	result.erg = GRAND_STREET;
    }
    else if (counter == SLOTS - 2)
    {
	g_print("SMALL STREET\n");
	result.erg = SMALL_STREET;
    }
}

gint
gamehandler_get_value(int move)
{
    int i;

    switch (move)
    {
#ifdef STRANGE_RULES
    case PASCH3:
	return 3 * (result.pasch & 7);
    case PASCH4:
	return 4 * (result.pasch & 7);
#endif
    case PASCH3:
    case PASCH4:
    {
	int tmp=0;
	for (i=0; i<6; i++)
	    tmp += result.numbers[i] * (i+1);
	return tmp;
    }
    case PASCH5:
	return PASCH5_VALUE;
    case FULL_HOUSE:
	return FULL_HOUSE_VALUE;
    case SMALL_STREET:
	return SMALL_STREET_VALUE;
    case GRAND_STREET:
	return GRAND_STREET_VALUE;		
    case CHANCE:
    {
	int tmp=0;
	for (i=0; i<6; i++)
	    tmp += result.numbers[i] * (i+1);
	return tmp;
    }
    default:
	return result.numbers[move] * (move+1);
    }
}

void
gamehandler_next_player()
{
    gamehandler.gameflow_status = GAMEFLOW_STARTTURN;
    
    if (++gamehandler.current_player >
	(gamehandler.player_num < MAX_PLAYERS ? gamehandler.player_num : MAX_PLAYERS))
	gamehandler.current_player = 1;
    
    g_print("NEXT PLAYER: %d\n", gamehandler.current_player);

    // Hilite the label for the new player
    table_hilite_player_label();

    if (gamehandler.player_num > 1)
	statusbar("It is the next players turn. Please roll the dice...");
    else
	statusbar("Please roll the dice...");
}

GtkWidget*
gamehandler_query_mainwidget()
{
    return mainwidget;
}
